/* $Id: compressutilsexample.c,v 1.2 1994/05/26 20:29:24 vertex Exp $ */

/* This is a simple program demonstrating how to use the compression routines
 * supplied by compressutils.lib. If you run this program without arguments,
 * it will load itself into a memory buffer, compress the data, decompress
 * it, and will compare the original data with the decompressed data to
 * make sure the compression process worked successfully. You can provide
 * a name on the command-line, and the program will load that file in and
 * compress it, instead of using the program's own binary for that purpose.
 *
 * Note that this program is essentially cutting corners in the way it manages
 * the callback function for the compressor. The buffer allocated to hold
 * the compressed data is twice as large as the actual file size of the file
 * being compressed. This is to allow for the eventuality of the compressed
 * data getting bigger than the original. This is highly unlikely to happen,
 * but it is possible. Smarter code would allocate a smaller buffer to hold
 * the compressed data. Once this smaller buffer gets full, it can get
 * dumped to a file.
 */

#include "types.h"
#include "filestream.h"
#include "filestreamfunctions.h"
#include "stdio.h"
#include "mem.h"
#include "compressutils.h"


/*****************************************************************************/


uint32 *originalData;
uint32  originalIndex;
uint32  originalMax;

uint32 *compressedData;
uint32  compressedIndex;
uint32  compressedMax;

uint32 *finalData;
uint32  finalIndex;
uint32  finalMax;


/*****************************************************************************/


void PutCompressedWord(void *context, uint32 word)
{
    if (compressedIndex < compressedMax)
        compressedData[compressedIndex++] = word;
}


/*****************************************************************************/


void PutFinalWord(void *context, uint32 word)
{
    if (finalIndex < finalMax)
        finalData[finalIndex++] = word;
}


/*****************************************************************************/


uint8 compWork[COMPRESSOR_BUFFER_SIZE];
uint8 decompWork[DECOMPRESSOR_BUFFER_SIZE];

int main(int argc, char **argv)
{
Stream       *stream;
Err           err;
bool          same;
uint32        i;
int32         fileSize;
char         *name;
Compressor   *comp;
Decompressor *decomp;

    if (argc > 2)
    {
        printf("Usage: compresstest [filename]\n(if no filename is given, then the program binary is used)");
        return (-1);
    }
    else if (argc == 1)
    {
        name = argv[0];
    }
    else
    {
        name = argv[1];
    }

    stream = OpenDiskStream(name,0);
    if (stream)
    {
        fileSize = stream->st_FileLength & 0xfffffffc;

        originalData   = (uint32 *)malloc(fileSize);
        compressedData = (uint32 *)malloc(fileSize*2);
        finalData      = (uint32 *)malloc(fileSize);

        if (originalData && compressedData && finalData)
        {
            if (ReadDiskStream(stream,(char *)originalData,fileSize) == fileSize)
            {
                originalIndex   = 0;
                originalMax     = fileSize / sizeof(uint32);
                compressedIndex = 0;
                compressedMax   = (fileSize*2) / sizeof(uint32);
                finalIndex      = 0;
                finalMax        = fileSize / sizeof(uint32);

                comp = CreateCompressor(compWork,PutCompressedWord,NULL,0);
                if (comp)
                {
                    FeedCompressor(comp,originalData,originalMax);
                    err = DeleteCompressor(comp);

                    if (err >= 0)
                    {
                        decomp = CreateDecompressor(decompWork,PutFinalWord,NULL,0);
                        if (decomp)
                        {
                            err = FeedDecompressor(decomp,compressedData,compressedIndex);
                            DeleteDecompressor(decomp);

                            if (err >= 0)
                            {
                                same = TRUE;
                                for (i = 0; i < originalMax; i++)
                                {
                                    if (originalData[i] != finalData[i])
                                    {
                                        same = FALSE;
                                        break;
                                    }
                                }

                                if (same)
                                {
                                    printf("Uncompressed data matched original\n");
                                }
                                else
                                {
                                    printf("Uncompressed data differed with original!\n");
                                    for (i = 0; i < 10; i++)
                                    {
                                        printf("orig $%08x, final $%08x, comp $%08x\n",
                                               originalData[i],
                                               finalData[i],
                                               compressedData[i]);
                                    }
                                }
                            }
                            else
                            {
                                printf("Error decompressing: %ld\n",err);
                            }
                        }
                    }
                    else
                    {
                        printf("Error compressing: %ld\n",err);
                    }
                }
            }
            else
            {
                printf("Could not read whole file\n");
            }
        }
        else
        {
            printf("Could not allocate memory buffers\n");
        }

        free(originalData);
        free(compressedData);
        free(finalData);

        CloseDiskStream(stream);
    }
    else
    {
        printf("Could not open '%s'\n",name);
    }

    return (0);
}
